package com.uas.ps.sync.consumer;

import com.alibaba.fastjson.JSONObject;
import com.uas.ps.core.util.ArrayUtils;
import com.uas.ps.core.util.Assert;
import com.uas.ps.core.util.DateFormatUtils;
import com.uas.ps.core.util.StringUtils;
import com.uas.ps.sync.entity.SyncMessage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.kafka.KafkaProperties;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.config.ConcurrentKafkaListenerContainerFactory;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.stereotype.Component;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

/**
 * @author sunyj
 * @since 2018/1/14 17:34
 */
@Component
public class KafkaConsumer {

    private static final String LOG_DIR = System.getProperty("java.io.tmpdir");
    private static final String LOG_FILE_PATH = (LOG_DIR.endsWith(File.separator) ? LOG_DIR : LOG_DIR + "/") + "consumer-update-error.log";
    private final Logger logger = LoggerFactory.getLogger(KafkaConsumer.class);
    private final JdbcTemplate jdbcTemplate;
    private final KafkaTemplate<String, String> kafkaTemplate;
    private final KafkaProperties kafkaProperties;
    private ConcurrentMap<String, Integer> batchCodeCount = new ConcurrentHashMap<>();

    @Autowired
    public KafkaConsumer(JdbcTemplate jdbcTemplate, KafkaTemplate<String, String> kafkaTemplate, ConcurrentKafkaListenerContainerFactory<String, SyncMessage> kafkaListenerContainerFactory, KafkaProperties kafkaProperties) {
        this.jdbcTemplate = jdbcTemplate;
        this.kafkaTemplate = kafkaTemplate;
        kafkaListenerContainerFactory.setBatchListener(true);
        this.kafkaProperties = kafkaProperties;
    }

    @KafkaListener(topics = "${kafka.consumer.topics}")
    public void listen(List<SyncMessage> messages) {
        try {
            logger.info("records size: " + messages.size());
            Map<String, ArgsMessage> operations = new HashMap<>();
            for (SyncMessage message : messages) {
                MessageResult messageResult;
                try {
                    messageResult = processMessage(message);
                } catch (Exception e) {
                    logger.error("处理失败", e);
                    continue;
                }
                String sql = messageResult.getSql();
                ArgsMessage argsMessage = operations.get(sql);
                if (argsMessage == null) {
                    argsMessage = new ArgsMessage();
                    operations.put(sql, argsMessage);
                }
                argsMessage.add(messageResult.getArgs(), message);
            }

            Set<Map.Entry<String, ArgsMessage>> entries = operations.entrySet();
            for (Map.Entry<String, ArgsMessage> entry : entries) {
                String sql = entry.getKey();
                ArgsMessage argsMessage = entry.getValue();
                List<Object[]> argsList = argsMessage.getArgsList();
                List<SyncMessage> messageList = argsMessage.getMessages();
                logger.info("batchUpdate... " + argsList.size());
                logger.info("sql... " + sql);
                logger.info("argsList... " + JSONObject.toJSONString(argsList));
                try {
                    // 先进行批量更新
                    int[] results = jdbcTemplate.batchUpdate(sql, argsList);
                    logger.info("results... " + Arrays.toString(results));
                } catch (Exception e) {
                    // 批量更新失败，进行单条更新
                    logger.error("批量更新失败", e);
                    logger.info("尝试单条更新");
                    for (int i = 0; i < argsList.size(); i++) {
                        Object[] args = argsList.get(i);
                        try {
                            int result = jdbcTemplate.update(sql, args);
                            logger.info("results... " + result);
                        } catch (Exception e1) {
                            // 单条更新失败，单独记录日志
                            logger.error("单条更新失败：" + messageList.get(i), e1);
                            try (FileWriter fileWriter = new FileWriter(LOG_FILE_PATH, true)) {
                                fileWriter.write(DateFormatUtils.DATETIME_FORMAT.format(new Date()) + " " + messageList.get(i) + "\n");
                                fileWriter.flush();
                            } catch (IOException e2) {
                                logger.error("日志写入失败", e2);
                            }
                        }
                    }
                }
            }
        } catch (Exception e) {
            logger.error("消息处理出错", e);
        }
    }

    private MessageResult processMessage(SyncMessage syncMessage) {
        logger.info("receiving message... " + syncMessage.toString());
        String batchCode = syncMessage.getBatchCode();
        String sourceApp = syncMessage.getSourceApp();
        String groupId = kafkaProperties.getConsumer().getGroupId();
        // batchCode 不为空并且来源应用与消费者的 group id 相同时，需要生成回复消息，以告知数据同步已经完成（只剩一个消费者，暂时去掉这个限制）
        if (!StringUtils.isEmpty(batchCode)) {
            Integer count = batchCodeCount.get(batchCode);
            if (count == null) {
                count = 1;
            } else {
                count++;
            }
            logger.info("consuming... " + count + " for batch " + batchCode);
            if (count.intValue() == syncMessage.getBatchSize()) {
                logger.info(String.format("batch %s processed: %d, sending message...", batchCode, count));
                kafkaTemplate.sendDefault(batchCode);
                logger.info(String.format("message sent: batchCode=%s, count=%d", batchCode, count));
            } else {
                batchCodeCount.put(batchCode, count);
            }
        }
        String methodType = syncMessage.getMethodType();
        Assert.hasText(methodType, "methodType is empty: " + methodType);
        // 根据不同的更改类型执行不同的操作
        switch (methodType.toLowerCase()) {
            case "delete":
                return delete(syncMessage);
            case "update":
                return update(syncMessage);
            case "insert":
                return insert(syncMessage);
            default:
                throw new IllegalArgumentException("Unsupported method type: " + methodType);
        }
    }

    /**
     * 根据消息删除相应的记录
     */
    private MessageResult delete(SyncMessage syncMessage) {
        String tableName = syncMessage.getTableName();
        JSONObject dataKey = syncMessage.getDataKey();
        Assert.hasText(tableName, "tableName is empty: " + tableName);
        Assert.notEmpty(dataKey, "dataKey is empty: " + dataKey);
        logger.info("deleting... table=" + tableName + ", key=" + dataKey);

        // 构造 sql
        // delete from ${tableName} where key3 = ? and key4 = ?
        StringBuilder sqlBuilder = new StringBuilder("delete from ");
        sqlBuilder.append(tableName);

        Object[] args = where(sqlBuilder, dataKey);
        return new MessageResult(sqlBuilder.toString(), args);
    }

    /**
     * 根据消息更新相应的记录
     */
    private MessageResult update(SyncMessage syncMessage) {
        String tableName = syncMessage.getTableName();
        JSONObject dataKey = syncMessage.getDataKey();
        JSONObject data = syncMessage.getData();
        Assert.hasText(tableName, "tableName is empty: " + tableName);
        Assert.notEmpty(dataKey, "dataKey is empty: " + dataKey);
        Assert.notEmpty(data, "data is empty: " + data);
        logger.info("updating... table=" + tableName + ", key=" + dataKey);

        // 构造 sql
        // update ${tableName} set key1 = ?, key2 = ? where key3 = ? and key4 = ?
        StringBuilder sqlBuilder = new StringBuilder("update ");
        sqlBuilder.append(tableName).append(" set ");

        Object[] updateArgs = connect(sqlBuilder, data, ", ", false);
        Object[] whereArgs = where(sqlBuilder, dataKey);
        return new MessageResult(sqlBuilder.toString(), ArrayUtils.concat(updateArgs, whereArgs));
    }

    /**
     * 根据消息插入相应的记录
     */
    private MessageResult insert(SyncMessage syncMessage) {
        String tableName = syncMessage.getTableName();
        JSONObject dataKey = syncMessage.getDataKey();
        JSONObject data = syncMessage.getData();
        Assert.hasText(tableName, "tableName is empty: " + tableName);
        Assert.notEmpty(dataKey, "dataKey is empty: " + dataKey);
        Assert.notEmpty(data, "data is empty: " + data);
        logger.info("inserting... table=" + tableName);

        // 构造 sql
        // insert into ${tableName} (key1, key2) values(?, ?)
        StringBuilder sqlBuilder = new StringBuilder("insert into ");
        sqlBuilder.append(tableName).append(" (");
        StringBuilder valuesBuilder = new StringBuilder(" values(");
        JSONObject map = dataKey;
        map.putAll(data);
        Object[] args = new Object[map.size()];
        Iterator<Map.Entry<String, Object>> iterator = map.entrySet().iterator();
        for (int i = 0; i < map.size(); i++) {
            Map.Entry<String, Object> entry = iterator.next();
            Object value = entry.getValue();
            if (i != 0) {
                sqlBuilder.append(", ");
                valuesBuilder.append(", ");
            }
            sqlBuilder.append(entry.getKey());
            valuesBuilder.append("?");
            args[i] = mayConvertToDate(value);
        }
        sqlBuilder.append(")");
        valuesBuilder.append(")");
        sqlBuilder.append(valuesBuilder);
        return new MessageResult(sqlBuilder.toString(), args);
    }

    /**
     * 构造 where 条件
     *
     * @param sqlBuilder 未构造完成的 sql
     * @param map        条件
     * @return 修改后的 sql 中的参数（${sqlBuilder} where key1 = ? and key2 =?）
     */
    private Object[] where(StringBuilder sqlBuilder, JSONObject map) {
        sqlBuilder.append(" where ");
        // 用作 where 中的值不可为 null
        return connect(sqlBuilder, map, " and ", true);
    }

    /**
     * 使用指定的分隔符连接键值对
     *
     * @param sqlBuilder    未构造完成的 sql
     * @param map           键值对
     * @param separator     分隔符
     * @param assertNotNull 是否检查键值对中的值不为 null
     * @return 修改后的 sql 中的参数（${sqlBuilder} key1 = ?${separator}key2 =?）
     */
    private Object[] connect(StringBuilder sqlBuilder, JSONObject map, String separator, boolean assertNotNull) {
        Object[] args = new Object[map.size()];
        Iterator<Map.Entry<String, Object>> iterator = map.entrySet().iterator();
        for (int i = 0; i < map.size(); i++) {
            Map.Entry<String, Object> entry = iterator.next();
            Object value = entry.getValue();
            if (i != 0) {
                sqlBuilder.append(separator);
            }
            sqlBuilder.append(entry.getKey()).append(" = ?");
            if (assertNotNull) {
                Assert.notNull(value, "value is null: " + entry);
            }
            args[i] = mayConvertToDate(value);
        }
        return args;
    }


    /**
     * 转换日期类型的值
     */
    private Object mayConvertToDate(Object value) {
        // 数据源为 oracle 时，将日期字符串转为日期类型
        // TODO isOracle
        // if (dataSource.isOracle() && value != null) {
        if (value != null) {
            try {
                String str = value.toString();
                if (DateFormatUtils.isDate(str)) {
                    return DateFormatUtils.DATE_FORMAT.parse(str);
                } else if (DateFormatUtils.isDateTime(str)) {
                    return DateFormatUtils.DATETIME_FORMAT.parse(str);
                }
            } catch (ParseException e) {
                throw new IllegalStateException("日期转换出错：" + value);
            }
        }
        return value;
    }

    private class MessageResult {
        private String sql;
        private Object[] args;

        private MessageResult(String sql, Object[] args) {
            this.sql = sql;
            this.args = args;
        }

        private String getSql() {
            return sql;
        }

        private Object[] getArgs() {
            return args;
        }
    }

    private class ArgsMessage {
        private List<Object[]> argsList = new ArrayList<>();
        private List<SyncMessage> messages = new ArrayList<>();

        private void add(Object[] args, SyncMessage message) {
            argsList.add(args);
            messages.add(message);
        }

        private List<Object[]> getArgsList() {
            return argsList;
        }

        private List<SyncMessage> getMessages() {
            return messages;
        }
    }
}
