package com.uas.ps.sync.core;

import com.uas.ps.core.util.Assert;
import com.uas.ps.sync.core.annotation.Sync;
import com.uas.ps.sync.core.util.ReflectUtils;
import org.aspectj.lang.ProceedingJoinPoint;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.jpa.repository.support.JpaEntityInformation;
import org.springframework.data.jpa.repository.support.SimpleJpaRepository;
import org.springframework.data.repository.CrudRepository;
import org.springframework.kafka.core.KafkaTemplate;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import java.io.Serializable;
import java.util.Arrays;

/**
 * 消息生产者
 *
 * @author sunyj
 * @since 2018/1/26 10:48
 */
public class MessageProducer {

    private final Logger logger = LoggerFactory.getLogger(getClass());

    private KafkaTemplate kafkaTemplate;

    @PersistenceContext
    private EntityManager entityManager;

    public MessageProducer(KafkaTemplate kafkaTemplate) {
        Assert.notNull(kafkaTemplate, "kafkaTemplate is null");
        this.kafkaTemplate = kafkaTemplate;
    }

    public void addSaveMessage(Object result) {
        if (result != null && result.getClass().isAnnotationPresent(Sync.class)) {
            if (result instanceof Iterable) {
                Iterable iterable = (Iterable) result;
                for (Object data : iterable) {
                    save(data);
                }
            } else {
                save(result);
            }
        }
    }

    public void addDeleteMessage(ProceedingJoinPoint proceedingJoinPoint) {
        logger.info("aroundDeleting: " + proceedingJoinPoint);
        Object proxy = proceedingJoinPoint.getThis();
        Object targetObject = null;
        try {
            targetObject = ReflectUtils.getTargetObject(proxy);
        } catch (Exception e) {
            e.printStackTrace();
        }
        System.out.println(targetObject instanceof CrudRepository);
        if (targetObject instanceof SimpleJpaRepository) {
            SimpleJpaRepository<?, ? extends Serializable> simpleJpaRepository = (SimpleJpaRepository<?, ? extends Serializable>) targetObject;
            JpaEntityInformation<?, ? extends Serializable> entityInformation = (JpaEntityInformation<?, ? extends Serializable>) ReflectUtils.getValue("entityInformation", simpleJpaRepository);

            Class<?> domainClass = entityInformation.getJavaType();
            Class<? extends Serializable> idType = entityInformation.getIdType();

            if (domainClass.isAnnotationPresent(Sync.class)) {
                Object[] args = proceedingJoinPoint.getArgs();
                if (args != null) {
                    System.out.println(Arrays.toString(args));
                    if (args.length == 0) {
                        System.out.println("deleteAll ...");
                    } else if (args.length == 1) {
                        Object arg = args[0];
                        if (arg.getClass() == idType) {
                            // deleteById can't cascade publish delete event,
                            // so first findOne(id), then use deleteEntity to cascade publish delete event
                            // Object e = simpleJpaRepository.findOne((Serializable)arg); // CrudMethodMetadata metadata is advised, and call simpleJpaRepository's method will skip normal setter step, and this will cause NullPointException
                            Object e = findOne(domainClass, arg);
                            if (e != null) {
                                delete(e);
                            }
                        } else if (arg.getClass() == domainClass) {
                            delete(arg);
                        } else if (arg instanceof Iterable) {
                            Iterable iterable = (Iterable) arg;
                            for (Object data : iterable) {
                                delete(data);
                            }
                        }
                    } else {
                        throw new IllegalStateException("The intercept of delete method in repository can support most one parameter. Currently we only support intercept method in " + CrudRepository.class);
                    }
                }
            }
        }
    }

    private Object findOne(Class<?> domainClass, Object id) {
        return entityManager.find(domainClass, id);
    }

    private void save(Object data) {
        logger.info("afterSaved: " + data);
        SyncMessage message = new SyncMessage(data.getClass(), SyncMessage.Type.SAVE, data);
        kafkaTemplate.sendDefault(message);
        logger.info("sent: " + message);
    }

    private void delete(Object data) {
        SyncMessage message = new SyncMessage(data.getClass(), SyncMessage.Type.DELETE, data);
        kafkaTemplate.sendDefault(message);
        logger.info("sent: " + message);
    }
}
