package com.uas.ps.sync.producer.jms;

import com.alibaba.fastjson.JSONObject;
import com.uas.ps.core.util.Assert;
import com.uas.ps.core.util.CollectionUtils;
import com.uas.ps.core.util.StringUtils;
import com.uas.ps.sync.entity.SyncMessage;
import com.uas.ps.sync.producer.schedule.Executable;
import com.uas.ps.sync.producer.schedule.TaskInformation;
import com.uas.ps.sync.producer.schedule.TaskService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.stereotype.Service;

import java.util.List;

/**
 * 对数据库的消息队列进行实时监听
 *
 * @author sunyj
 * @since 2018/1/14 16:29
 */
@Service
public class JmsListener {

    /**
     * 第一次执行的延迟时间间隔为 3 秒，以防止日志交杂错乱，不好查看
     */
    private static final long INITIAL_DELAY = 3000;
    /**
     * 两次任务之间的等待时间间隔为 1 秒钟
     */
    private static final long INTERVAL = 1000;

    @Autowired
    private SyncMessageDao syncMessageDao;

    @Autowired
    private TaskService taskService;

    @Autowired
    private SyncMessageService syncMessageService;

    @Autowired
    private KafkaTemplate<String, SyncMessage> kafkaTemplate;

    private TaskInformation taskInformation;

    private Logger logger = LoggerFactory.getLogger(getClass());

    /**
     * 开启监听
     *
     * @param interval 每次接收到jms消息后等待的时间（秒）
     * @return 开启成功与否的提示信息
     */
    public void start(Long interval) {
        if (interval == null) {
            interval = INTERVAL;
        } else {
            interval *= 1000;
        }
        if (interval <= 0) {
            throw new IllegalArgumentException("interval 不合法：" + interval);
        }

        if (isRunning()) {
            throw new IllegalStateException("已存在运行的监听服务");
        }

        try {
            String title = "监听消息";
            Executable command = new Executable() {
                @Override
                public String execute() {
                    SPage<SyncMessage> sPage = syncMessageService.findAll(1, 100);
                    List<SyncMessage> syncMessages = sPage.getContent();
                    if (CollectionUtils.isEmpty(syncMessages)) {
                        return "无消息";
                    }
                    for (SyncMessage syncMessage : syncMessages) {
                        try {
                            process(syncMessage);
                        } catch (Exception e) {
                            logger.error("消息处理失败", e);
                        }
                    }
                    // 如果消息不止一页，立即消费下一页
                    if (sPage.getTotalPage() > 1) {
                        execute();
                    }
                    return "正常";
                }
            };
            taskInformation = new TaskInformation(title, command, INITIAL_DELAY, interval, TaskInformation.ScheduleType.FixedDelay);
            taskService.newTask(taskInformation);
            if (!taskService.isStopped()) {
                taskService.stop();
            }
            taskService.start();
        } catch (Exception e) {
            if (taskInformation != null) {
                taskInformation = null;
            }
            throw new IllegalStateException("开启失败", e);
        }
    }

    /**
     * 关闭监听
     *
     * @return 关闭成功与否的提示信息
     */
    public void stop() {
        if (!isRunning()) {
            throw new IllegalStateException("监听服务未开启或已关闭");
        } else {
            taskService.remove(taskInformation.getCode());
            taskInformation = null;
            if (!taskService.isStopped()) {
                taskService.stop();
            }
            taskService.start();
        }
    }

    /**
     * @return 监听服务是否正在运行
     */
    public boolean isRunning() {
        return taskInformation != null && taskService.exist(taskInformation.getCode());
    }

    /**
     * 将消息推到 kafka
     *
     * @param syncMessage 消息
     */
    private void process(SyncMessage syncMessage) {
        logger.info("Sending message... " + syncMessage.toString());
        Long id = syncMessage.getId();
        syncMessageDao.markRetryCount(id);

        JSONObject dataKey = syncMessage.getDataKey();
        JSONObject data = syncMessage.getData();
        String methodType = syncMessage.getMethodType();
        // 根据不同的更改类型对字段进行检查
        switch (methodType.toLowerCase()) {
            case "delete":
                Assert.notEmpty(dataKey, "dataKey is empty: " + dataKey);
                break;
            case "update":
            case "insert":
                Assert.notEmpty(dataKey, "dataKey is empty: " + dataKey);
                Assert.notEmpty(data, "data is empty: " + data);
                break;
            default:
                throw new IllegalArgumentException("Unsupported method type: " + methodType);
        }

        // 发送消息
        String batchCode = syncMessage.getBatchCode();
        // 同一个 batchCode 的消息发送到同一分区
        if (!StringUtils.isEmpty(batchCode)) {
            kafkaTemplate.sendDefault(batchCode, syncMessage);
        } else {
            kafkaTemplate.sendDefault(syncMessage);
        }
        syncMessageDao.dequeueMessage(id);
    }
}
